#!/usr/bin/env lua53

local sqlite3 = require("lsqlite3");

local openssl = require("openssl");
      openssl.rand = require("openssl.rand");
      openssl.digest = require("openssl.digest");

local crypto = {};
local cgi = {};
local html = {};
      html.board = {};
      html.post = {};
      html.container = {};
      html.table = {};
      html.list = {};
      html.pdp = {};
      html.string = {};
local board = {};
local post = {};
local image = {};
local identity = {};
      identity.session = {};
local log = {};
local global = {};
      global.announce = {};

math.randomseed(os.time());

local nanodb = sqlite3.open("nanochan.db");

-- Ensure all required tables exist.
nanodb:exec("CREATE TABLE IF NOT EXISTS Global (Announce)");
nanodb:exec("CREATE TABLE IF NOT EXISTS Boards (Name, Title, Subtitle, MaxPostNumber, Lock, AllowImages, RepostImages)");
nanodb:exec("CREATE TABLE IF NOT EXISTS Posts (Board, Number, Parent, Date, LastBumpDate, Name, Email, Subject, Comment, Image, Sticky, Cycle, Autosage, Lock)");
nanodb:exec("CREATE TABLE IF NOT EXISTS Accounts (Name, Type, Board, PwHash, Salt)");
nanodb:exec("CREATE TABLE IF NOT EXISTS Sessions (Key, Account, ExpireDate)");
nanodb:exec("CREATE TABLE IF NOT EXISTS Logs (Name, Board, Date, Description)");
nanodb:busy_timeout(1000);

--
-- Miscellaneous functions.
--

function string.tokenize(input, delimiter)
    local result = {};

    if input == nil then
        return {};
    end

    for match in (input .. delimiter):gmatch("(.-)" .. delimiter) do
        table.insert(result, match);
    end

    return result;
end

-- This function is used to generate salts for passwords, etc.
-- The output of this function must be cryptographically secure, and therefore it uses
-- openssl.rand.uniform() instead of math.random();
function string.random(length, pattern)
    local length = length or 64;
    local pattern = pattern or "a-zA-Z0-9"
    local result = "";
    local ascii = "";
    local dict = "";

    for i = 0, 255 do
        ascii = ascii .. string.char(i);
    end

    dict = ascii:gsub("[^" .. pattern .. "]", "");

    while string.len(result) < length do
        local randidx = openssl.rand.uniform(string.len(dict)) + 1;
        local randbyte = dict:byte(randidx);
        result = result .. string.char(randbyte);
    end

    return result;
end

function string.striphtml(input)
    local result = input;
    result = result:gsub("&", "&amp;");
    result = result:gsub("<", "&lt;");
    result = result:gsub(">", "&gt;");
    result = result:gsub("\"", "&quot;");
    result = result:gsub("'", "&#39;");
    return result;
end

function io.exists(filename)
    local f = io.open(filename, "r");

    if f ~= nil then
        f:close();
        return true;
    else
        return false;
    end
end

--
-- Global announcement functions.
--

function global.announce.retrieve()
    for tbl in nanodb:nrows("SELECT Announce FROM Global") do
        return tbl["Announce"];
    end

    nanodb:exec("INSERT INTO Global VALUES ('')");
    return "";
end

function global.announce.update(newstring)
    if global.announce.retrieve() == nil then
        nanodb:exec("INSERT INTO Global VALUES ('')");
    end

    local stmt = nanodb:prepare("UPDATE Global SET Announce = ?");
    stmt:bind_values(newstring);
    stmt:step();
    stmt:finalize();
end

--
-- Cryptographic functions.
--

function crypto.hash(hashtype, data)
    local bstring = openssl.digest.new(hashtype):final(data);
    local result = "";

    for i = 1, #bstring do
        result = result .. string.format("%02x", string.byte(bstring:sub(i,i)));
    end

    return result;
end

--
-- Board-related functions.
--

-- Get a list of all boards, return it as a table. The table contains board names.
function board.list()
    local boards = {}
    local i = 1;

    for tbl in nanodb:nrows("SELECT Name FROM Boards ORDER BY MaxPostNumber DESC") do
        boards[i] = tbl["Name"];
        i = i + 1;
    end

    return boards;
end

-- Get all board information into a table. The table is read-only and any values written
-- to it will not be reflected in the database.
function board.retrieve(name)
    local stmt = nanodb:prepare("SELECT * FROM Boards WHERE Name = ?");
    stmt:bind_values(name);

    for row in stmt:nrows() do
        stmt:finalize();
        return row;
    end

    return nil;
end

function board.validname(name)
    return name and ((not name:match("[^a-z0-9]")) and (#name > 0) and (#name <= 8));
end

function board.validtitle(title)
    return title and ((#title > 0) and (#title <= 32));
end

function board.validsubtitle(subtitle)
    return subtitle and ((#subtitle >= 0) and (#subtitle <= 64));
end

-- Check if a board exists. Return 'true' if it does, and 'false' if it does not.
function board.exists(name)
    if not board.validname(name) then
        return false;
    elseif board.retrieve(name) == nil then
        return false;
    else
        return true;
    end
end

-- Format a board name (place slashes around it).
function board.format(name)
    return board.validname(name) and ("/" .. name .. "/") or nil;
end

-- Create a board. There are three required pieces of information: board name, board
-- title (short description) and board subtitle.
function board.create(name, title, subtitle)
    if not board.validname(name) then
        return nil;
    end

    local stmt = nanodb:prepare("INSERT INTO Boards VALUES (?,?,?,?,?,?,?)");

    local maxpostnumber = 0;
    local lock = 0;
    local allowimages = 1;
    local repostimages = 1;

    stmt:bind_values(name,
                     string.striphtml(title),
                     string.striphtml(subtitle),
                     maxpostnumber,
                     lock,
                     allowimages,
                     repostimages);
    stmt:step();
    stmt:finalize();
end

-- Change board settings.
function board.update(board_tbl)
    local stmt = nanodb:prepare("UPDATE Boards SET " ..
                                "Title = ?, Subtitle = ?, Lock = ?, " ..
                                "AllowImages = ?, RepostImages = ? " ..
                                "WHERE Name = ?");

    stmt:bind_values(string.striphtml(board_tbl["Title"]), string.striphtml(board_tbl["Subtitle"]),
                     board_tbl["Lock"], board_tbl["AllowImages"], board_tbl["RepostImages"], board_tbl["Name"]);
    stmt:step();
    stmt:finalize();
end

-- Delete a board.
function board.delete(name)
    if not board.validname(name) then
        return nil;
    end

    nanodb:exec("DELETE FROM Boards WHERE Name = '" .. name .. "'");
    nanodb:exec("DELETE FROM Posts WHERE Board = '" .. name .. "'");
    nanodb:exec("DELETE FROM Accounts WHERE Board = '" .. name .. "'");
end

--
-- Identity (account) functions.
--

function identity.list()
    local identities = {};
    local i = 1;

    for tbl in nanodb:nrows("SELECT Name FROM Accounts ORDER BY Name") do
        identities[i] = tbl["Name"];
        i = i + 1;
    end

    return identities;
end

function identity.retrieve(name)
    local stmt = nanodb:prepare("SELECT * FROM Accounts WHERE Name = ?");
    stmt:bind_values(name);

    for row in stmt:nrows() do
        stmt:finalize();
        return row;
    end

    return nil
end

function identity.exists(name)
    if identity.validname(name) == nil then
        return false;
    end

    if identity.retrieve(name) == nil then
        return false;
    else
        return true;
    end
end

-- Class can be either:
--   * "admin" - Site administrator, unlimited powers
--   * "bo" - Board owner, powers limited to a single board
--   * "gvol" - Global volunteer, powers limited by site administrators
--   * "lvol" - Local volunteer, powers limited by board owners, powers limited to a single board
function identity.create(class, name, password, board)
    local board = board or "Global";
    local stmt = nanodb:prepare("INSERT INTO Accounts VALUES (?,?,?,?,?)");

    -- Generate a random salt.
    local salt = string.random(64);

    -- SHA-512 hash the password, with the salt appended to the end of the password.
    local hash = crypto.hash("sha512", password .. salt);

    stmt:bind_values(name, class, board, hash, salt);
    stmt:step();
    stmt:finalize();
end

function identity.validname(name)
    return (not name:match("[^a-zA-Z0-9]")) and (#name >= 1) and (#name <= 16);
end

function identity.delete(name)
    if not identity.validname(name) then
        return nil;
    end

    nanodb:exec("DELETE FROM Accounts WHERE Name = '" .. name .. "'");
    nanodb:exec("DELETE FROM Sessions WHERE Account = '" .. name .. "'");
    nanodb:exec("UPDATE Logs SET Name = '<i>Deleted</i>' WHERE Name = '" .. name .. "'");
end

function identity.changepassword(name, password)
    local stmt = nanodb:prepare("UPDATE Accounts SET PwHash = ?, Salt = ? WHERE Name = ?");
    local salt = string.random(64);
    local hash = crypto.hash("sha512", password .. salt);
    stmt:bind_values(hash, salt, name);
    stmt:step();
    stmt:finalize();
end

function identity.validpassword(password)
    return (#password >= 6) and (#password <= 64);
end

function identity.validclass(class)
    return (class == "admin" or
            class == "gvol" or
            class == "bo" or
            class == "lvol")
end

function identity.valid(name, password)
    if not (identity.validname(name) and identity.validpassword(password)) then
        return false;
    end

    local identity_tbl = identity.retrieve(name);

    if identity_tbl == nil then
        return false;
    end

    if crypto.hash("sha512", password .. identity_tbl["Salt"]) == identity_tbl["PwHash"] then
        return true;
    else
        return false;
    end
end

-- Delete all session keys from a particular user (e.g. logout).
function identity.session.delete(user)
    local stmt = nanodb:prepare("DELETE FROM Sessions WHERE Account = ?");
    stmt:bind_values(user);
    stmt:step();
    stmt:finalize();
end

-- Create a new session key for this user and add it to the database.
function identity.session.create(user)
    -- Clear any existing keys for this user to prevent duplicates.
    identity.session.delete(user);

    local key = string.random(32);
    local expiry = os.time() + 3600; -- key expires in 1 hour

    local stmt = nanodb:prepare("INSERT INTO Sessions VALUES (?,?,?)");

    stmt:bind_values(key, user, expiry);
    stmt:step();
    stmt:finalize();

    return key;
end

-- Reset the expiry time of the specified session key to 3600 seconds (1 hour) from
-- the current time.
function identity.session.refresh(user)
    local stmt = nanodb:prepare("UPDATE Sessions SET ExpireDate = ? WHERE Account = ?");
    stmt:bind_values(os.time() + 3600, user);
    stmt:step();
    stmt:finalize();
end

-- Return the user for which the specified session key is valid. Return nil if the
-- session key is not valid.
function identity.session.valid(key)
    local result = nil;
    if key == nil then return nil end;

    for tbl in nanodb:nrows("SELECT * FROM Sessions") do
        if os.time() > tbl["ExpireDate"] then
            -- Clean away any expired session keys.
            identity.session.delete(tbl["Account"]);
        elseif tbl["Key"] == key then
            result = tbl["Account"];
        end
    end

    identity.session.refresh(result);
    return result;
end

--
-- Image handling functions.
--

-- Detect the format of an image (PNG, JPG, GIF).
function image.format(data)
    if data:sub(1,8) == "\x89PNG\x0D\x0A\x1A\x0A" then
        return "png";
    elseif data:sub(1,3) == "\xFF\xD8\xFF" then
        return "jpg";
    elseif data:sub(1,6) == "GIF87a"
        or data:sub(1,6) == "GIF89a" then
        return "gif";
    else
        return "???";
    end
end

function image.pathname(filename)
    return "media/" .. filename;
end

function image.thumbnail(filename)
    return "media/thumb/" .. filename;
end

function image.icon(filename)
    return "media/icon/" .. filename;
end

function image.exists(filename)
    if filename == nil or filename == "" then
        return false;
    end

    if io.exists(image.pathname(filename)) and
       io.exists(image.thumbnail(filename)) then
        return true;
    else
        return false;
    end
end

-- Create a thumbnail which will fit into a 200x200 grid.
-- Imagemagick (convert) must be installed for this to work.
-- Will not modify images which are smaller than 200x200.
function image.create_thumbnail(filename)
    local path_orig = image.pathname(filename);
    local path_thumb = image.thumbnail(filename);
    return os.execute("convert " .. path_orig .. " -resize 200x200\\> " .. path_thumb);
end

-- Create a catalog icon (even smaller than a normal thumbnail).
function image.create_icon(filename)
    local path_orig = image.pathname(filename);
    local path_icon = image.icon(filename);
    return os.execute("convert " .. path_orig .. " -resize 100x70\\> " .. path_icon);
end

-- Save an image and return its hashed filename. Errors will result in returning nil.
-- Image hashes are always SHA-256 for compatibility with 8chan and friends.
function image.save(data, create_catalog_icon)
    local hash = crypto.hash("sha256", data);
    local format = image.format(data);
    if format == "???" then return nil end;
    local filename = hash .. "." .. format;
    local path = image.pathname(filename);

    if io.exists(path) then
        -- No need to process and save the image again, since it has already been saved
        -- previously.
        return filename;
    end;

    local imagefile = io.open(path, "w");
    imagefile:write(data);
    imagefile:close();

    image.create_thumbnail(filename);

    if create_catalog_icon == true then
        image.create_icon(filename);
    end

    return filename;
end

-- Delete an image.
function image.delete(filename)
    os.remove(image.pathname(filename));
    os.remove(image.thumbnail(filename));
    os.remove(image.icon(filename));
end

--
-- Post functions.
--

function post.retrieve(boardname, number)
    local stmt = nanodb:prepare("SELECT * FROM Posts WHERE Board = ? AND Number = ?");
    stmt:bind_values(boardname, tonumber(number));

    for row in stmt:nrows() do
        stmt:finalize();
        return row;
    end

    return nil;
end

function post.listthreads(boardname)
    local stmt = nanodb:prepare("SELECT Number FROM Posts WHERE Board = ? AND Parent = 0 AND Sticky = 1 ORDER BY LastBumpDate DESC");
    local threads = {};
    local i = 1;
    stmt:bind_values(boardname);

    for tbl in stmt:nrows() do
        threads[i] = tonumber(tbl["Number"]);
        i = i + 1;
    end

    stmt = nanodb:prepare("SELECT Number FROM Posts WHERE Board = ? AND Parent = 0 AND Sticky = 0 ORDER BY LastBumpDate DESC");
    stmt:bind_values(boardname);

    for tbl in stmt:nrows() do
        threads[i] = tonumber(tbl["Number"]);
        i = i + 1;
    end

    return threads;
end

function post.exists(boardname, number)
    if post.retrieve(boardname, number) == nil then
        return false;
    else
        return true;
    end
end

function post.bump(boardname, number)
    local stmt = nanodb:prepare("UPDATE Posts SET LastBumpDate = ? WHERE Board = ? AND Number = ?");
    stmt:bind_values(os.time(), boardname, tonumber(number));
    stmt:step();
    stmt:finalize();
end

function post.toggle(attribute, boardname, number)
    local current_value = post.retrieve(boardname, number)[attribute];
    local new_value = (current_value == 1) and 0 or 1;
    local stmt = nanodb:prepare("UPDATE Posts SET " .. attribute .. " = ? WHERE Board = ? AND Number = ?");
    stmt:bind_values(new_value, boardname, number);
    stmt:step();
    stmt:finalize();
end

function post.threadreplies(boardname, number)
    local replies = {};
    local i = 1;

    for tbl in nanodb:nrows("SELECT Number FROM Posts WHERE Board = '" .. boardname .. "' AND Parent = " .. number) do
        replies[i] = tonumber(tbl["Number"]);
        i = i + 1;
    end

    return replies;
end

function post.format(boardname, number)
    return board.format(boardname) .. number;
end

-- Turn nanochan-formatting into html.
function post.nano2html(text)
    local text = "\n" .. text .. "\n";
    text = text:gsub("&gt;&gt;(%d+)", "<a class='reference' href='#post%1'>&gt;&gt;%1</a>");
    text = text:gsub("&gt;&gt;&gt;/([%d%l]-)/", "<a class='reference' href='/nano/%1'>&gt;&gt;&gt;/%1/</a>");
    text = text:gsub("(https?://)(.-)%s", "<a href='%1%2'>%1%2</a> ");
    text = text:gsub("\n&gt;(.-)\n", "\n<span class='greentext'>&gt;%1</span>\n");
    text = text:gsub("\n&lt;(.-)\n", "\n<span class='pinktext'>&lt;%1</span>\n");
    text = text:gsub("%(%(%((.-)%)%)%)", "<span class='kiketext'>(((%1)))</span>");
    text = text:gsub("==(.-)==", "<span class='redtext'>%1</span>");
    text = text:gsub("%*%*(.-)%*%*", "<span class='spoiler'>%1</span>");
    text = text:gsub("~~(.-)~~", "<s>%1</s>");
    text = text:gsub("__(.-)__", "<u>%1</u>");
    text = text:gsub("&#39;&#39;&#39;(.-)&#39;&#39;&#39;", "<b>%1</b>");
    text = text:gsub("&#39;&#39;(.-)&#39;&#39;", "<i>%1</i>");
    text = text:gsub("\n", "<br />");
    return text;
end

-- This function does not delete the actual file. It simply removes the reference to that
-- file, and deletes the file if there are no remaining references.
function post.unlink(boardname, number)
    local post_tbl = post.retrieve(boardname, number);
    local filename = post_tbl["Image"];

    if filename == "" then
        return;
    end

    local stmt = nanodb:prepare("UPDATE Posts SET Image = '' WHERE Board = ? AND Number = ?");
    stmt:bind_values(boardname, number);
    stmt:step();
    stmt:finalize();

    local stmt = nanodb:prepare("SELECT COUNT(Number) FROM Posts WHERE Image = ?");
    stmt:bind_values(filename);
    stmt:step();
    local occurrences = stmt:get_value(0);
    stmt:finalize();

    if occurrences == 0 then
        image.delete(filename);
    end
end

function post.delete(boardname, number)
    local post_tbl = post.retrieve(boardname, number);

    post.unlink(boardname, number);
    local stmt = nanodb:prepare("DELETE FROM Posts WHERE Board = ? AND Number = ?");
    stmt:bind_values(boardname, number);
    stmt:step();
    stmt:finalize();

    if post_tbl["Parent"] == 0 then
        local replies = post.threadreplies(boardname, number);

        for i = 1, #replies do
            post.delete(boardname, replies[i]);
        end
    end
end

function post.create(skey, boardname, parent, name, email, subject, comment, file)
    local number = board.retrieve(boardname)["MaxPostNumber"] + 1;
    local parent = parent or 0;
    local name = (name and name ~= "") and string.striphtml(name) or "Anonymous";
    local email = email and string.striphtml(email) or "";
    local subject = subject and string.striphtml(subject) or "";
    local comment = comment and post.nano2html(string.striphtml(comment)) or "";
    local file = file or "";
    local date = os.time();
    local lastbumpdate = (parent == 0 and os.time() or 0);
    local autosage = (parent == 0 and (email == "sage" and 1 or 0) or 0);
    local parent_tbl = post.retrieve(boardname, parent);

    if parent ~= 0 and #post.threadreplies(boardname, parent) >= 350 and parent_tbl["Cycle"] == 1 then
        local stmt = nanodb:prepare("DELETE FROM Posts WHERE Number = (SELECT Number FROM Posts WHERE Parent = ? AND Board = ? ORDER BY Date LIMIT 1)");
        stmt:bind_values(parent, boardname);
        stmt:step();
        stmt:finalize();
    elseif parent == 0 and #post.listthreads(boardname) >= 350 then
        local stmt = nanodb:prepare("DELETE FROM Posts WHERE Number = (SELECT Number FROM Posts WHERE Parent = 0 AND Sticky = 0 AND Board = ? ORDER BY Date LIMIT 1)");
        stmt:bind_values(boardname);
        stmt:step();
        stmt:finalize();
    end

    if name == "##" then
        -- Implement capcode.
        local username = identity.session.valid(skey);

        if username ~= nil then
            local acct_tbl = identity.retrieve(username);
            local acctclass = acct_tbl["Type"];
            local assignboard = acct_tbl["Board"];
            local capcode;

            if acctclass == "admin" then
                capcode = "Nanochan Administrator";
            elseif acctclass == "bo" then
                capcode = "Board Owner (" .. board.format(assignboard) .. ")";
            elseif acctclass == "gvol" then
                capcode = "Global Volunteer";
            elseif acctclass == "lvol" then
                capcode = "Board Volunteer (" .. board.format(assignboard) .. ")";
            end

            name = username .. " <span class='capcode'>## " .. capcode .. "</span>";
        else
            name = "Anonymous";
        end
    end

    local stmt = nanodb:prepare("INSERT INTO Posts VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)");
    stmt:bind_values(boardname, number, parent, date, lastbumpdate, name, email, subject, comment, file, 0, 0, autosage, 0);
    stmt:step();
    stmt:finalize();

    stmt = nanodb:prepare("UPDATE Boards SET MaxPostNumber = ? WHERE Name = ?");
    stmt:bind_values(number, boardname);
    stmt:step();
    stmt:finalize();

    if parent ~= 0 then
        if not ((parent_tbl["Autosage"] == 1) or (string.lower(email) == "sage")) then
            post.bump(boardname, parent);
        end
    end

    return number;
end

--
-- Log access functions.
--

function log.create(desc, account, boardname)
    local account = account or "<i>System</i>";
    local board = board.format(boardname) or "<i>Global</i>";
    local date = os.time();
    local stmt = nanodb:prepare("INSERT INTO Logs VALUES (?,?,?,?)");

    stmt:bind_values(account, board, date, desc);
    stmt:step();
    stmt:finalize();
end

function log.retrieve(limit, offset)
    local limit = limit or 64;
    local offset = offset or 0;
    local entries = {};
    local i = 1;

    for tbl in nanodb:nrows("SELECT * FROM Logs ORDER BY Date DESC LIMIT " .. limit .. " OFFSET " .. offset) do
        entries[i] = tbl;
        i = i + 1;
    end

    return entries;
end

--
-- HTML output functions.
--

function html.redirect(location)
    io.write("<!DOCTYPE html>\n");
    io.write("<html>");
    io.write(  "<head>");
    io.write(    "<title>Redirecting...</title>");
    io.write(    "<meta http-equiv='refresh' content='0; url=", location or "/", "' />");
    io.write(  "</head>");
    io.write("</html>");
end

function html.begin(title, name, value)
    if title == nil then
        title = ""
    else
        title = title .. " - "
    end

    io.write("<!DOCTYPE html>\n");
    io.write("<html>");
    io.write(  "<head>");
    io.write(    "<title>", title, "nanochan</title>");
    io.write(    "<link rel='stylesheet' type='text/css' href='/static/nanochan.css' />");

    if name and value then
        io.write("<meta http-equiv='set-cookie' content='", name, "=", value, ";Path=/nano' />");
    end

    io.write(    "<meta charset='utf-8' />");
    io.write(    "<meta name='viewport' content='width=device-width, initial-scale=1.0' />");
    io.write(  "</head>");
    io.write(  "<body>");
    io.write(    "<div id='topbar'>");
    io.write(      "<nav id='topnav'>");
    io.write(        "<ul>");
    io.write(          "<li class='system'><a href='/nano'>main</a></li>");
    io.write(          "<li class='system'><a href='/nano/System/mod'>mod</a></li>");
    io.write(          "<li class='system'><a href='/nano/System/log'>log</a></li>");

    local boards = board.list();
    for i = 1, #boards do
        io.write("<li class='board'><a href='/nano/", boards[i], "'>", board.format(boards[i]), "</a></li>");
    end

    io.write(        "</ul>");
    io.write(      "</nav>");
    io.write(    "</div>");
    io.write(    "<div id='content'>");
end

function html.finish()
    io.write(    "</div>");
    io.write(  "</body>");
    io.write("</html>");
end

function html.redheader(text)
    io.write("<h1 class='redheader'>", text, "</h1>");
end

function html.announce(text)
    io.write("<div id='announce'>", text, "</div>");
end

function html.container.begin(type)
    io.write("<div class='container ", type or "narrow", "'>");
end

function html.container.finish()
    io.write("</div>");
end

function html.container.barheader(text)
    io.write("<h2 class='barheader'>", text, "</h2>");
end

function html.table.begin(...)
    local arg = {...};

    io.write("<table>");
    io.write("<tr>");

    for i = 1, #arg do
        io.write("<th>", arg[i], "</th>");
    end

    io.write("</tr>");
end

function html.table.entry(...)
    local arg = {...};

    io.write("<tr>");

    for i = 1, #arg do
        io.write("<td>", tostring(arg[i]), "</td>");
    end

    io.write("</tr>");
end

function html.table.finish()
    io.write("</table>");
end

function html.list.begin(type)
    io.write(type == "ordered" and "<ol>" or "<ul>");
end

function html.list.entry(text, class)
    io.write("<li", class and (" class='" .. class .. "' ") or "", ">", text, "</li>");
end

function html.list.finish(type)
    io.write(type == "ordered" and "</ol>" or "</ul>");
end

-- Pre-defined pages.
function html.pdp.authorization_denied()
    html.begin("permission denied");
    html.redheader("Permission denied");
    html.container.begin();
    io.write("Your account class lacks authorization to perform this action. <a href='/nano/System/mod'>Go back.</a>");
    html.container.finish();
    html.finish();
end

function html.pdp.error(heading, explanation)
    html.begin("error");
    html.redheader(heading);
    html.container.begin();
    io.write(explanation);
    html.container.finish();
    html.finish();
end

function html.pdp.notfound()
    html.begin("404");
    html.redheader("404 Not Found");
    html.container.begin();
    io.write("The resource which was requested does not appear to exist. Please check the URL");
    io.write(" and try again. Alternatively, if you believe this error message to in itself");
    io.write(" be an error, try contacting the nanochan administration.");
    html.container.finish();
    html.finish();
end

function html.string.link(href, text, title)
    local result = "<a href='" .. href .. "'";

    if href:sub(1, 1) ~= "/" then
        result = result .. " rel='noreferrer' target='_blank'";
    end

    if title then
        result = result .. " title='" .. title .. "'";
    end

    result = result .. ">" .. (text or href) .. "</a>";
    return result;
end

function html.string.datetime(unixtime)
    return "<time datetime='" .. os.date("!%Y-%m-%d %H:%M:SS", unixtime) .. "'>" .. os.date("!%d/%m/%Y %H:%M:%S", unixtime) .. "</time>";
end

function html.string.boardlink(boardname)
    return html.string.link("/nano/" .. boardname, board.format(boardname));
end

function html.board.title(boardname)
    io.write("<h1 id='boardtitle'>", board.format(boardname), " - ", board.retrieve(boardname)["Title"], "</h1>");
end

function html.board.subtitle(boardname)
    io.write("<h2 id='boardsubtitle'>", board.retrieve(boardname)["Subtitle"], "</h2>");
end

function html.post.postform(boardname, parent)
    io.write("<fieldset><form id='postform' action='/nano/System/post' method='post' enctype='multipart/form-data'>");
    io.write("<input type='hidden' name='board' value='" .. boardname .. "' />");
    io.write("<input type='hidden' name='parent' value='" .. parent .. "' />");
    io.write("<label for='name'>Name</label><input type='text' id='name' name='name' /><br />");
    io.write("<label for='email'>Email</label><input type='text' id='email' name='email' /><br />");
    io.write("<label for='subject'>Subject</label><input type='text' id='subject' name='subject' autocomplete='off' />");
    io.write("<input type='submit' value='Post' /><br />");
    io.write("<label for='comment'>Comment</label><textarea id='comment' name='comment' form='postform' rows='5' cols='35'></textarea><br />");
    io.write("<label for='file'>File</label><input type='file' id='file' name='file' />");
    io.write("</form></fieldset>");
end

function html.post.catalog(boardname)
    local threadlist = post.listthreads(boardname);

    io.write("<a href='/nano/", boardname, "'>[Update]</a>");
    io.write("<hr />");
    io.write("<div class='catalog-container'>");

    for i = 1, #threadlist do
        local number = threadlist[i];
        local post_tbl = post.retrieve(boardname, number);

        io.write("<div class='catalog-thread'>");
        io.write(  "<div class='catalog-thread-link'><a href='/nano/", boardname, "/", number, "'>");

        if image.exists(post_tbl["Image"]) then
            io.write("<img src='/media/icon/" .. post_tbl["Image"] .. "' />");
        else
            io.write("***");
        end

        io.write(  "</a></div>");
        io.write(  "<div class='thread-info'>");
        io.write(    "<span class='thread-info-replies'>R:", #post.threadreplies(boardname, number), "</span>");
        io.write(    "<span class='thread-info-flags'>");
        if post_tbl["Sticky"] == 1 then io.write("(S)"); end;
        if post_tbl["Lock"] == 1 then io.write("(L)"); end;
        if post_tbl["Autosage"] == 1 then io.write("(A)"); end;
        if post_tbl["Cycle"] == 1 then io.write("(C)"); end;
        io.write(    "</span>");
        io.write(  "</div>");
        io.write(  "<div class='thread-mod-links'>");
        io.write(    "<a href='/nano/System/mod/post/delete/", boardname, "/", number, "' title='Delete'>[D]</a>");

        if image.exists(post_tbl["Image"]) then
            io.write("<a href='/nano/System/mod/post/unlink/", boardname, "/", number, "' title='Unlink File'>[U]</a>");
            io.write("<a href='/nano/System/mod/file/delete/", post_tbl["Image"], "' title='Delete File'>[F]</a>");
        end

        io.write(    "<a href='/nano/System/mod/post/sticky/", boardname, "/", number, "' title='Sticky'>[S]</a>");
        io.write(    "<a href='/nano/System/mod/post/lock/", boardname, "/", number, "' title='Lock'>[L]</a>");
        io.write(    "<a href='/nano/System/mod/post/autosage/", boardname, "/", number, "' title='Autosage'>[A]</a>");
        io.write(    "<a href='/nano/System/mod/post/cycle/", boardname, "/", number, "' title='Cycle'>[C]</a>");
        io.write(  "</div>");
        io.write(  "<div class='catalog-thread-subject'>");
        io.write(     post_tbl["Subject"] or "");
        io.write(  "</div>");
        io.write(  "<div class='catalog-thread-comment'>");
        io.write(     post_tbl["Comment"]);
        io.write(  "</div>");
        io.write("</div>");
    end

    io.write("</div>");
end

function html.post.render(boardname, number)
    local post_tbl = post.retrieve(boardname, number);

    io.write("<div class='post' id='post", number, "'>");
    io.write(  "<div class='post-header'>");

    io.write(    "<span class='post-subject'>", post_tbl["Subject"] or "", "</span>");
    io.write(    "<span class='post-name'>");

    if post_tbl["Email"] ~= "" then
        io.write("<a class='post-email' href='mailto:", post_tbl["Email"], "'>");
    end

    io.write(    post_tbl["Name"]);

    if post_tbl["Email"] ~= "" then
        io.write("</a>");
    end

    io.write(    "</span>");

    io.write(    "<span class='post-date'>", html.string.datetime(post_tbl["Date"]), "</span>");
    io.write(    "<span class='post-number'>No.", post_tbl["Number"], "</span>");

    if post_tbl["Parent"] == 0 then
        io.write("<span class='thread-info-flags'>");
        if post_tbl["Sticky"] == 1 then io.write("(S)"); end;
        if post_tbl["Lock"] == 1 then io.write("(L)"); end;
        if post_tbl["Autosage"] == 1 then io.write("(A)"); end;
        if post_tbl["Cycle"] == 1 then io.write("(C)"); end;
        io.write("</span>");
    end

    io.write(    "<span class='post-mod-links'>");
    io.write(      "<a href='/nano/System/mod/post/delete/", boardname, "/", number, "' title='Delete'>[D]</a>");

    if image.exists(post_tbl["Image"]) then
        io.write(  "<a href='/nano/System/mod/post/unlink/", boardname, "/", number, "' title='Unlink File'>[U]</a>");
        io.write(  "<a href='/nano/System/mod/file/delete/", post_tbl["Image"], "' title='Delete File'>[F]</a>");
    end

    if post_tbl["Parent"] == 0 then
        io.write(  "<a href='/nano/System/mod/post/sticky/", boardname, "/", number, "' title='Sticky'>[S]</a>");
        io.write(  "<a href='/nano/System/mod/post/lock/", boardname, "/", number, "' title='Lock'>[L]</a>");
        io.write(  "<a href='/nano/System/mod/post/autosage/", boardname, "/", number, "' title='Autosage'>[A]</a>");
        io.write(  "<a href='/nano/System/mod/post/cycle/", boardname, "/", number, "' title='Cycle'>[C]</a>");
    end

    io.write(    "</span>");

    io.write(  "</div>");

    if image.exists(post_tbl["Image"]) then
        io.write("<div class='post-image-info'>");
        io.write("Image: <a href='/media/" .. post_tbl["Image"] .. "' download>" .. post_tbl["Image"] .. "</a>");
        io.write("</div>");
        io.write("<a target='_blank' href='/media/" .. post_tbl["Image"] .. "'>");
        io.write("<img class='post-image' src='/media/thumb/", post_tbl["Image"], "' />");
        io.write("</a>");
    end

    io.write(  "<div class='post-comment'>");
    io.write(  post_tbl["Comment"]);
    io.write(  "</div>");
    io.write("</div>");

    io.write("<br />");
end

function html.post.renderthread(boardname, number)
    local replies = post.threadreplies(boardname, number);
    html.post.render(boardname, number);

    for i = 1, #replies do
        html.post.render(boardname, replies[i]);
    end
end

--
-- CGI- and HTTP-related functions
--

-- Decode a string in the format: This%20is%20a%20string%0Acontaining%20a%20newline
function cgi.decode_percentencoding(input)
    return (input:gsub("+", " "):gsub("%%(%x%x)", function(xx) return string.char(tonumber(xx, 16)) end));
end

-- Decode a string in the format: param1=value1&param2=value2&param3=value3
function cgi.decode_urlencoding(input)
    if input == nil then
        return {}
    end

    local kvpairs = string.tokenize(input, "&");
    local parameters = {};

    for i = 1, #kvpairs do
        local kvpair = string.tokenize(kvpairs[i], "=");

        if #kvpair == 2 then
            parameters[kvpair[1]] = cgi.decode_percentencoding(kvpair[2]);
        end
    end

    return parameters;
end

function cgi.decode_multipart(input)
    if input == nil then
        return {}
    end

    local boundary = string.tokenize(string.tokenize(os.getenv("CONTENT_TYPE"), "; ")[2], "=")[2];
    local segments = string.tokenize(input, boundary);
    table.remove(segments, 1);
    table.remove(segments, #segments);
    local finaltable = {};

    for i = 1, #segments do
        local header_lines = string.tokenize(segments[i]:gsub("\r", ""), "\n\n")[1];
        local name = header_lines:match("name=\"(.-)\"");
        local value = segments[i]:match("\r\n\r\n(.-)\r\n$");
        finaltable[name] = value;
    end

    return finaltable;
end

-- Initialize cgi variables.
cgi.pathinfo = os.getenv("PATH_INFO") and string.tokenize(os.getenv("PATH_INFO"):gsub("^/", ""), "/") or {};
cgi.referer = os.getenv("HTTP_REFERER");
cgi.usragent = os.getenv("HTTP_USER_AGENT");
cgi.reqmethod = os.getenv("REQUEST_METHOD");
cgi.contentlen = os.getenv("CONTENT_LENGTH");
cgi.get = cgi.decode_urlencoding(os.getenv("QUERY_STRING"));
cgi.cookie = cgi.decode_urlencoding(os.getenv("HTTP_COOKIE"));

if cgi.reqmethod == "POST" then
    if os.getenv("CONTENT_TYPE") == "application/x-www-form-urlencoded" then
        cgi.post = cgi.decode_urlencoding(io.read(tonumber(cgi.contentlen)));
    else
        cgi.post = cgi.decode_multipart(io.read(tonumber(cgi.contentlen)));
    end
else
    cgi.post = {};
end

-- Write HTTP headers.
io.write("Content-Type: text/html; charset=utf-8\n");
io.write("Cache-Control: no-cache\n");
io.write("\n");

--
-- This is the main part of Nanochan, where all the pages are defined.
--

local skey = cgi.cookie["session_key"];
local username = identity.session.valid(skey);
local acctclass = username and identity.retrieve(username)["Type"] or nil;
local assignboard = username and identity.retrieve(username)["Board"] or nil;

if cgi.pathinfo[1] == nil then
    -- /nano
    html.begin();
    html.redheader("Welcome to Nanochan");
    html.container.begin("narrow");
    io.write("Nanochan is an imageboard, similar to 4cancer or 8cuck - only better. :^)");
    html.container.barheader("Boards");

    io.write("Boards are ordered from most posts to least posts.");

    local boards = board.list();
    html.list.begin("ordered");
    for i = 1, #boards do
        local board_tbl = board.retrieve(boards[i]);
        html.list.entry(html.string.boardlink(board_tbl["Name"]) .. " - " .. board_tbl["Title"]);
    end
    html.list.finish("ordered");

    html.container.barheader("Rules");
    io.write("These rules apply to all boards on nanochan:");
    html.list.begin("ordered");
    html.list.entry("Content which is illegal in the country of hosting will be removed. " ..
                    "Typically, this means child porn, bomb threats/mass shooting threats, etc.");
    html.list.entry("Flooding is not permitted. We define flooding as posting similar posts more " ..
                    "than 5 times per hour, or posting in such a way that it significantly " .. 
                    "changes the composition of a board. Common sense will be utilized.");
    html.list.finish("ordered");
    io.write("Individual boards may set their own rules which apply to that board. However, note");
    io.write(" that the nanochan rules stated above apply to everything done on the website.");

    html.container.barheader("Administration");
    io.write("To contact the administration, send an e-mail to this address: ", html.string.link("mailto:37564N@memeware.net"));

    html.container.finish();
    html.finish();
elseif cgi.pathinfo[1] == "System" then
    -- /nano/System/,,,
    if cgi.pathinfo[2] == nil then
        -- /nano/System
        html.redirect("/nano");
        os.exit();
    elseif cgi.pathinfo[2] == "log" then
        -- /nano/System/log/...
        html.begin("logs");
        html.redheader("Nanochan Log");
        html.container.begin("wide");
        html.table.begin("Account", "Board", "Time", "Description");

        local entries = log.retrieve();
        for i = 1, #entries do
            html.table.entry(entries[i]["Name"],
                             entries[i]["Board"],
                             html.string.datetime(entries[i]["Date"]),
                             entries[i]["Description"]);
        end

        html.table.finish();
        html.container.finish();
        html.finish();
        os.exit();
    elseif cgi.pathinfo[2] == "mod" then
        -- /nano/System/mod/...
        if cgi.pathinfo[3] == "login" then
            -- /nano/System/mod/login
            -- This area is the only area in /nano/System/mod which unauthenticated users are
            -- allowed to access.
            if cgi.post["username"] and cgi.post["password"] then
                if #identity.list() == 0 then
                    -- Special case: if there are no mod accounts, use the first supplied credentials to
                    -- establish an administration account (to allow for board creation and the like).
                    identity.create("admin", cgi.post["username"], cgi.post["password"]);
                    log.create("Created a new admin account for board Global: " .. cgi.post["username"]);
                    html.redirect("/nano/System/mod/login");
                else
                    -- User has supplied a username and a password. Check if valid.
                    if identity.valid(cgi.post["username"], cgi.post["password"]) then
                        -- Set authentication cookie.
                        html.begin("successful login", "session_key", identity.session.create(cgi.post["username"]));
                        html.redheader("Login successful");
                        html.container.begin();
                        io.write("You have successfully logged in. You may now ", html.string.link("/nano/System/mod", "continue"), " to the moderation tools.");
                        html.container.finish();
                        html.finish();
                    else
                        html.begin("invalid credentials");
                        html.redheader("Error");
                        html.container.begin();
                        io.write("Either your username, your password, or both your username and your");
                        io.write(" password were invalid. Please ", html.string.link("/nano/System/mod/login", "return"));
                        io.write(" and try again.");
                        html.container.finish();
                        html.finish();
                    end
                end

                os.exit();
            end

            html.begin("moderation");
            html.redheader("Moderator login");
            html.container.begin();
            io.write("The moderation tools require a login. Access to moderation tools is restricted");
            io.write(" to administrators, global volunteers, board owners and board volunteers.");

            if #identity.list() == 0 then
                io.write("<br /><b>There are currently no moderator accounts. As such, the credentials you");
                io.write(" type in the box below will become those of the first administrator account.</b>");
            end

            html.container.barheader("Login");
            io.write("<fieldset><form method='post'>");
            io.write(  "<label for='username'>Username</label><input type='text' id='username' name='username' /><br />");
            io.write(  "<label for='password'>Password</label><input type='password' id='password' name='password' /><br />");
            io.write(  "<label for='submit'>Submit</label><input id='submit' type='submit' value='Continue' />");
            io.write("</form></fieldset>");
            html.container.finish();
            html.finish();
            os.exit();
        end

        if username == nil then
            -- The user does not have a valid session key. User must log in.
            html.redirect("/nano/System/mod/login");
            os.exit();
        end

        if cgi.pathinfo[3] == nil then
            -- /nano/System/mod
            html.begin("moderation");
            html.redheader("Moderation Tools");
            html.container.begin();
            io.write("<a id='logout-button' href='/nano/System/mod/logout'>[Logout]</a>");
            io.write("You are logged in as <b>", username, "</b>.");
            io.write(" Your account class is <b>", acctclass, "</b>.");

            if acctclass == "bo" or acctclass == "lvol" then
                io.write("<br />You are assigned to <b>", html.string.boardlink(assignboard), "</b></a>.");
            end

            if acctclass == "admin" then
                html.container.barheader("Global");
                html.list.begin("unordered");
                html.list.entry(html.string.link("/nano/System/mod/global/announce", "Change top-bar announcement"));
                html.list.finish("unordered");
            end

            if acctclass == "admin" or acctclass == "bo" then
                html.container.barheader("Boards");
                html.list.begin("unordered");

                if acctclass == "admin" then
                    html.list.entry(html.string.link("/nano/System/mod/board/create", "Create a board"));
                    html.list.entry(html.string.link("/nano/System/mod/board/delete", "Delete a board"));
                end

                if acctclass == "admin" then
                    html.list.entry(html.string.link("/nano/System/mod/board/config", "Configure a board"));
                elseif acctclass == "bo" then
                    html.list.entry(html.string.link("/nano/System/mod/board/config/" .. assignboard, "Configure your board"));
                end

                html.list.finish("unordered");
            end

            html.container.barheader("Accounts");
            html.list.begin("unordered");

            if acctclass == "admin" or acctclass == "bo" then
                html.list.entry(html.string.link("/nano/System/mod/account/create", "Create an account"));
                html.list.entry(html.string.link("/nano/System/mod/account/delete", "Delete an account"));
                html.list.entry(html.string.link("/nano/System/mod/account/config", "Configure an account"));
            end

            html.list.entry(html.string.link("/nano/System/mod/account/config/" .. username, "Account settings"));
            html.list.finish("unordered");
            html.container.finish();
            html.finish();
        elseif cgi.pathinfo[3] == "logout" then
            identity.session.delete(username);
            html.redirect("/nano/System/mod/login");
        elseif cgi.pathinfo[3] == "board" then
            -- /nano/System/mod/board/...
            
            if cgi.pathinfo[4] == "create" then
                if acctclass ~= "admin" then
                    html.pdp.authorization_denied();
                    os.exit();
                end

                -- /nano/System/mod/board/create
                html.begin("create board");
                html.redheader("Create a board");
                html.container.begin();

                if cgi.post["board"] and cgi.post["title"] then
                    if not cgi.post["subtitle"] then
                        cgi.post["subtitle"] = "";
                    end

                    if board.exists(cgi.post["board"]) then
                        io.write("That board already exists.");
                    elseif not board.validname(cgi.post["board"]) then
                        io.write("Invalid board name.");
                    elseif not board.validtitle(cgi.post["title"]) then
                        io.write("Invalid board title.");
                    elseif not board.validsubtitle(cgi.post["subtitle"]) then
                        io.write("Invalid board subtitle.");
                    else
                        board.create(cgi.post["board"],
                                     cgi.post["title"],
                                     cgi.post["subtitle"]);
                        log.create("Created a new board: " .. board.format(cgi.post["board"]), username);
                        io.write("Board created: ", html.string.boardlink(cgi.post["board"]));
                    end
                end

                html.container.barheader("Instructions");
                html.list.begin("unordered");
                html.list.entry("<b>Board names</b> must consist of only lowercase characters and" ..
                                " numerals. They must be from one to eight characters long.");
                html.list.entry("<b>Board titles</b> must be from one to 32 characters long.");
                html.list.entry("<b>Board subtitles</b> must be from zero to 64 characters long.");
                html.list.finish("unordered");

                html.container.barheader("Enter board information");
                io.write("<fieldset><form method='post'>");
                io.write(  "<label for='board'>Name</label><input type='text' id='board' name='board' /><br />");
                io.write(  "<label for='title'>Title</label><input type='text' id='title' name='title' /><br />");
                io.write(  "<label for='subtitle'>Subtitle</label><input type='text' id='subtitle' name='subtitle' /><br />");
                io.write(  "<label for='submit'>Submit</label><input type='submit' id='submit' value='Create' /><br />");
                io.write("</form></fieldset>");

                html.container.finish();
                html.finish();
            elseif cgi.pathinfo[4] == "delete" then
                -- /nano/System/mod/board/delete
                if acctclass ~= "admin" then
                    html.pdp.authorization_denied();
                    os.exit();
                end

                html.begin("delete board");
                html.redheader("Delete a board");
                html.container.begin();

                if cgi.post["board"] then
                    if not board.exists(cgi.post["board"]) then
                        io.write("The board you specified does not exist.");
                    else
                        board.delete(cgi.post["board"]);
                        log.create("Deleted board <b>" .. board.format(cgi.post["board"]) .. "</b>" ..
                                   (cgi.post["reason"] and (" with reason: " .. cgi.post["reason"]) or ""), username);
                        io.write("Board deleted.");
                    end
                end

                html.container.barheader("Instructions");
                io.write("Deleting a board removes the board itself, along with all posts on that board,");
                io.write(" and all accounts assigned to that board. Board deletion is irreversible.");

                html.container.barheader("Enter information");
                io.write("<fieldset><form method='post'>");
                io.write(  "<label for='board'>Board</label><input type='text' id='board' name='board' /><br />");
                io.write(  "<label for='reason'>Reason</label><input type='text' id='reason' name='reason' /><br />");
                io.write(  "<label for='submit'>Submit</label><input type='submit' id='submit' value='Delete' /><br />");
                io.write("</form></fieldset>");

                html.container.finish();
                html.finish();
            elseif cgi.pathinfo[4] == "config" then
                -- /nano/System/mod/board/config
                if acctclass ~= "admin" and acctclass ~= "bo" then
                    html.pdp.authorization_denied();
                    os.exit();
                end

                if cgi.post["board"] then
                    html.redirect("/nano/System/mod/board/config/" .. cgi.post["board"]);
                    os.exit();
                end

                html.begin("configure board");
                html.redheader("Configure " .. (cgi.pathinfo[5] and board.format(cgi.pathinfo[5]) or "a board"));
                html.container.begin();

                if cgi.pathinfo[5] then
                    -- /nano/System/mod/board/config/...
                    if not board.exists(cgi.pathinfo[5]) then
                        io.write("That board does not exist. ", html.string.link("/nano/System/mod/board/config", "Go back"));
                        io.write(" and try again.");
                    elseif acctclass == "bo" and cgi.pathinfo[5] ~= assignboard then
                        io.write("You are not assigned to that board and are unable to configure it.");
                    else
                        if cgi.post["action"] then
                            local new_settings = {
                                Name =           cgi.pathinfo[5],
                                Title =          cgi.post["title"] or "",
                                Subtitle =       cgi.post["subtitle"] or "",
                                Lock =          (cgi.post["lock"] ~= nil and 1 or 0),
                                AllowImages =   (cgi.post["allowimages"] ~= nil and 1 or 0),
                                RepostImages =  (cgi.post["repostimages"] ~= nil and 1 or 0)
                            };

                            board.update(new_settings);
                            log.create("Edited board settings", username, cgi.pathinfo[5]);
                            io.write("Board settings modified.");
                        end

                        local existing = board.retrieve(cgi.pathinfo[5]);

                        io.write("<fieldset><form method='post'>");
                        io.write(  "<input type='hidden' name='action' value='yes' />");
                        io.write(  "<label for='name'>Name</label><input id='name' name='name' type='text' value='" .. existing["Name"] .. "' disabled /><br />");
                        io.write(  "<label for='title'>Title</label><input id='title' name='title' type='text' value='" .. existing["Title"] .. "' /><br />");
                        io.write(  "<label for='subtitle'>Subtitle</label><input id='subtitle' name='subtitle' type='text' value='" .. existing["Subtitle"] .. "' /><br />");
                        io.write(  "<label for='lock'>Lock</label><input id='lock' name='lock' type='checkbox' " .. (existing["Lock"] == 0 and "" or "checked ") .. "/><br />");
                        io.write(  "<label for='allowimages'>Images</label><input id='allowimages' name='allowimages' type='checkbox' "
                                   .. (existing["AllowImages"] == 0 and "" or "checked ") .. "/><br />");
                        io.write(  "<label for=repostimages'>Image Repost</label><input id='repostimages' name='repostimages' type='checkbox' "
                                   .. (existing["RepostImages"] == 0 and "" or "checked ") .. "/><br />");
                        io.write(  "<label for='submit'>Submit</label><input id='submit' type='submit' value='Update' />");
                        io.write("</form></fieldset>");
                    end
                else
                    html.container.barheader("Enter information");
                    io.write("<fieldset><form method='post'>");
                    io.write(  "<label for='board'>Board</label><input type='text' id='board' name='board' /><br />");
                    io.write(  "<label for='submit'>Submit</label><input type='submit' id='submit' value='Configure' /><br />");
                    io.write("</form></fieldset>");
                end

                html.container.finish();
                html.finish();
            end
        elseif cgi.pathinfo[3] == "global" then
            -- /nano/System/mod/global
            if cgi.pathinfo[4] == "announce" then
                html.begin("edit global announcement");
                html.redheader("Edit global announcement");
                html.container.begin();

                if cgi.post["action"] ~= nil then
                    global.announce.update(cgi.post["announce"] or "");
                    log.create("Edited global announcement", username);
                    io.write("Global announcement updated.");
                end

                io.write("<fieldset><form method='post'>");
                io.write(  "<input type='hidden' name='action' value='yes' />");
                io.write(  "<label for='announce'>Announcement</label><input type='text' id='announce' name='announce' value='" ..
                            string.striphtml(global.announce.retrieve()) .. "' /><br />");
                io.write(  "<label for='submit'>Submit</label><input type='submit' id='submit' value='Update' />");
                io.write("</form></fieldset>");

                html.container.finish();
                html.finish();
            end
        elseif cgi.pathinfo[3] == "account" then
            -- /nano/System/mod/account/...
            if cgi.pathinfo[4] == "create" then
                -- /nano/System/mod/account/create

                if acctclass ~= "admin" and acctclass ~= "bo" then
                    html.pdp.authorization_denied();
                    os.exit();
                end

                html.begin("create account");
                html.redheader("Create an account");
                html.container.begin();

                if cgi.post["account"] and cgi.post["password"] then
                    if acctclass == "bo" then
                        cgi.post["class"] = "lvol";
                        cgi.post["board"] = assignboard;
                    elseif cgi.post["class"] == "gvol" or cgi.post["class"] == "admin" then
                        cgi.post["board"] = nil;
                    end

                    if identity.exists(cgi.post["account"]) then
                        io.write("That account already exists.");
                    elseif not identity.validname(cgi.post["account"]) then
                        io.write("Invalid account name.");
                    elseif not identity.validpassword(cgi.post["password"]) then
                        io.write("Invalid password.");
                    elseif not identity.validclass(cgi.post["class"]) then
                        io.write("Invalid account class.");
                    elseif cgi.post["board"] and not board.exists(cgi.post["board"]) then
                        io.write("Board does not exist.");
                    else
                        identity.create(cgi.post["class"],
                                        cgi.post["account"],
                                        cgi.post["password"],
                                        cgi.post["board"]);
                        log.create("Created a new " .. cgi.post["class"] .. " account for board " ..
                                   board.format(cgi.post["board"]) .. ": " .. cgi.post["account"], username);
                        io.write("Account created.");
                    end
                end

                html.container.barheader("Instructions");
                html.list.begin("unordered");
                html.list.entry("<b>Usernames</b> can only consist of alphanumerics. They must be from 1 to 16 characters long.");
                html.list.entry("<b>Passwords</b> must be from 6 to 64 characters long.");
                if acctclass == "admin" then
                    html.list.entry("An account's <b>board</b> has no effect for Global Volunteers and " ..
                                    "Administrators. For Board Owners and Board Volunteers, the board " ..
                                    "parameter defines the board in which that account can operate.");
                end
                html.list.finish("unordered");

                html.container.barheader("Enter account information");
                io.write("<fieldset><form id='acctinfo' method='post'>");
                if acctclass == "admin" then
                    io.write("<label for='class'>Type</label>");
                    io.write("<select id='class' name='class' form='acctinfo'>");
                    io.write(  "<option value='admin'>Administrator</option>");
                    io.write(  "<option value='gvol'>Global Volunteer</option>");
                    io.write(  "<option value='bo'>Board Owner</option>");
                    io.write(  "<option value='lvol'>Board Volunteer</option>");
                    io.write("</select><br />");
                    io.write("<label for='board'>Board</label><input type='text' id='board' name='board' /><br />");
                end
                io.write("<label for='account'>Username</label><input type='text' id='account' name='account' /><br />");
                io.write("<label for='password'>Password</label><input type='password' id='password' name='password' /><br />");
                io.write("<label for='submit'>Submit</label><input type='submit' id='submit' value='Create' /><br />");
                io.write("</form></fieldset>");

                html.container.finish();
                html.finish();
            elseif cgi.pathinfo[4] == "delete" then
                -- /nano/System/account/delete
                if acctclass ~= "admin" and acctclass ~= "bo" then
                    html.pdp.authorization_denied();
                    os.exit();
                end

                html.begin("delete account");
                html.redheader("Delete an account");
                html.container.begin();

                if cgi.post["account"] then
                    if not identity.exists(cgi.post["account"]) then
                        io.write("The account which you have specified does not exist.");
                    elseif acctclass == "bo" and identity.retrieve(cgi.post["account"])["Board"] ~= assignboard then
                        io.write("You are not authorized to delete that account.");
                    else
                        identity.delete(cgi.post["account"]);
                        log.create("Deleted account <b>" .. board.format(cgi.post["account"]) ..
                                   "</b>" .. (cgi.post["reason"] and (" with reason: " .. cgi.post["reason"]) or ""), username);
                        io.write("Account deleted.");
                    end
                end

                html.container.barheader("Instructions");
                html.list.begin("unordered");
                html.list.entry("Deleting an account will log the user out of all active sessions.");
                html.list.entry("Deleting an account will replace names all logs associated with that account with '<i>Deleted</i>'.");
                html.list.finish("unordered");

                html.container.barheader("Enter information");
                io.write("<fieldset><form method='post'>");
                io.write("<label for='account'>Username</label><input type='text' id='account' name='account' /><br />");
                io.write("<label for='reason'>Reason</label><input type='text' id='reason' name='reason' /><br />");
                io.write("<label for='submit'>Submit</label><input type='submit' id='submit' value='Delete' /><br />");
                io.write("</form></fieldset>");
                html.container.finish();
                html.finish();
            elseif cgi.pathinfo[4] == "config" then
                -- /nano/System/mod/account/config/...
                if cgi.post["account"] then
                    html.redirect("/nano/System/mod/account/config/" .. cgi.post["account"]);
                    os.exit();
                end

                if cgi.pathinfo[5] then
                    if acctclass ~= "admin" and cgi.pathinfo[5] ~= username then
                        html.pdp.authorization_denied();
                        os.exit();
                    elseif not identity.exists(cgi.pathinfo[5]) then
                        html.pdp.error("Account not found", "The account that you specified does not exist.");
                        os.exit();
                    end

                    html.begin("configure account");
                    html.redheader("Configure account " .. cgi.pathinfo[5]);
                    html.container.begin();

                    if cgi.post["password1"] and cgi.post["password2"] then
                        if cgi.post["password1"] ~= cgi.post["password2"] then
                            io.write("The two passwords did not match.");
                        elseif not identity.validpassword(cgi.post["password1"]) then
                            io.write("Invalid password.");
                        else
                            identity.changepassword(cgi.pathinfo[5], cgi.post["password1"]);
                            log.create("Changed password for account: " .. cgi.pathinfo[5], username);
                            io.write("Password changed.");
                        end
                    end

                    html.container.barheader("Instructions");
                    html.list.begin();
                    html.list.entry("<b>Passwords</b> must be from 6 to 64 characters long.");
                    html.list.finish();
                    html.container.barheader("Enter information");
                    io.write("<fieldset><form method='post'>");
                    io.write("<label for='password1'>New password</label><input type='password' id='password1' name='password1' /><br />");
                    io.write("<label for='password2'>Repeat</label><input type='password' id='password2' name='password2' /><br />");
                    io.write("<label for='submit'>Submit</label><input type='submit' id='submit' value='Change' /><br />");
                    io.write("</form></fieldset>");
                    html.container.finish();
                    html.finish();
                else
                    html.begin("configure account");
                    html.redheader("Configure an account");
                    html.container.begin();
                    html.container.barheader("Enter information");
                    io.write("<fieldset><form method='post'>");
                    io.write("<label for='account'>Username</label><input type='text' id='account' name='account' /><br />");
                    io.write("<label for='submit'>Submit</label><input type='submit' label='submit' value='Configure' /><br />");
                    io.write("</form></fieldset>");
                    html.container.finish();
                    html.finish();
                end
            end
        elseif cgi.pathinfo[3] == "file" then
            local filename = cgi.pathinfo[5];

            if acctclass ~= "admin" and acctclass ~= "gvol" then
                html.pdp.authorization_denied();
                os.exit();
            elseif not image.exists(filename) then
                html.pdp.error("Invalid file", "The file you are trying to modify does not exist.");
                os.exit();
            end

            if cgi.pathinfo[4] == "delete" then
                log.create("Deleted file " .. filename .. " from all boards", username);
                image.delete(filename);
            end

            html.redirect(cgi.referer);
        elseif cgi.pathinfo[3] == "post" then
            local boardname = cgi.pathinfo[5];
            local number = tonumber(cgi.pathinfo[6]);

            if not post.exists(boardname, number) then
                html.pdp.error("Invalid post", "The post you are trying to modify does not exist.");
                os.exit();
            elseif (acctclass == "bo" or acctclass == "lvol") and assignboard ~= boardname then
                html.pdp.authorization_denied();
                os.exit();
            end

            if cgi.pathinfo[4] == "sticky" then
                log.create("Stickied thread " .. post.format(boardname, number), username, boardname);
                post.toggle("Sticky", boardname, number);
            elseif cgi.pathinfo[4] == "lock" then
                log.create("Locked thread " .. post.format(boardname, number), username, boardname);
                post.toggle("Lock", boardname, number);
            elseif cgi.pathinfo[4] == "autosage" then
                log.create("Autosaged thread " .. post.format(boardname, number), username, boadname);
                post.toggle("Autosage", boardname, number);
            elseif cgi.pathinfo[4] == "cycle" then
                log.create("Cycled thread " .. post.format(boardname, number), username, boardname);
                post.toggle("Cycle", boardname, number);
            elseif cgi.pathinfo[4] == "delete" then
                log.create("Deleted post " .. post.format(boardname, number), username, boardname);
                post.delete(boardname, number);
            elseif cgi.pathinfo[4] == "unlink" then
                log.create("Unlinked file from post " .. post.format(boardname, number), username, boardname);
                post.unlink(boardname, number);
            end

            html.redirect(cgi.referer);
        end
    elseif cgi.pathinfo[2] == "post" then
        -- /nano/System/post
        local post_board = cgi.post["board"];
        local post_parent = tonumber(cgi.post["parent"]);
        local post_name = cgi.post["name"];
        local post_email = cgi.post["email"];
        local post_subject = cgi.post["subject"];
        local post_comment = cgi.post["comment"];
        local post_file = cgi.post["file"];
        local parent_tbl = post.retrieve(post_board, post_parent);
        local board_tbl = board.retrieve(post_board);

        if cgi.post["board"] and cgi.post["parent"] then
            if not board_tbl then
                html.pdp.error("Invalid board", "The board you tried to post to does not exist.");
                os.exit();
            elseif post_parent ~= 0 and not post.exists(post_board, post_parent) then
                html.pdp.error("Invalid thread", "The thread you tried to post in does not exist. Perhaps it has been deleted.");
                os.exit();
            elseif parent_tbl ~= nil and parent_tbl["Lock"] == 1 and username == nil then
                html.pdp.error("Thread locked", "The thread you tried to post in is currently locked.");
                os.exit();
            elseif board_tbl["Lock"] == 1 and username == nil then
                html.pdp.error("Board locked", "The board you tried to post in is currently locked.");
                os.exit();
            elseif post_parent ~= 0 and parent_tbl["Parent"] ~= 0 then
                html.pdp.error("Invalid thread", "The thread you tried to post in is not a thread. This is not supported.");
                os.exit();
            elseif post_file and #post_file > 8388608 then
                html.pdp.error("File too large", "The file attached to your post was over 8 MiB. Please use a smaller file.");
                os.exit();
            elseif post_comment and #post_comment > 32768 then
                html.pdp.error("Post too long", "Your post text was over 32 KiB. Please reduce its length.");
                os.exit();
            elseif post_comment and select(2, post_comment:gsub("\n", "")) > 128 then
                html.pdp.error("Too many newlines", "Your post contained over 128 newlines. Please reduce its length.");
                os.exit();
            elseif post_name and #post_name > 64 then
                html.pdp.error("Name too long", "The text in the name field was over 64 bytes. Please reduce its length.");
                os.exit();
            elseif post_email and #post_email > 64 then
                html.pdp.error("Email too long", "The text in the email field was over 64 bytes. Please reduce its length.");
                os.exit();
            elseif post_subject and #post_subject > 64 then
                html.pdp.error("Subject too long", "The text in the subject field was over 64 bytes. Please reduce its length.");
                os.exit();
            elseif post_parent ~= 0 and #post.threadreplies(post_board, post_parent) >= 350 then
                html.pdp.error("Thread full", "The thread you tried to post in already has 350 replies. Please start a new thread instead.");
                os.exit();
            end

            local post_filename = image.save(post_file, (post_parent == 0));

            if post_file ~= "" and post_filename == nil then
                html.pdp.error("File error", "There was an issue with the file you uploaded. Nanochan only supports PNG, JPG and GIF files.");
                os.exit();
            end

            local post_number = post.create(skey, post_board, post_parent, post_name, post_email, post_subject, post_comment, post_filename);

            if post_parent == 0 then
                -- Redirect to the newly created thread.
                html.redirect("/nano/" .. post_board .. "/" .. post_number);
            else
                -- Redirect to the parent thread, but scroll down to the newly created post.
                html.redirect("/nano/" .. post_board .. "/" .. post_parent .. "#post" .. post_number);
            end
        end
    end
else
    -- /nano/b/,..
    if not board.exists(cgi.pathinfo[1]) then
        html.pdp.notfound();
        os.exit();
    end

    local board_tbl = board.retrieve(cgi.pathinfo[1]);
    html.begin(board.format(board_tbl["Name"]));
    html.board.title(board_tbl["Name"]);
    html.board.subtitle(board_tbl["Name"]);
    html.announce(global.announce.retrieve());

    if cgi.pathinfo[2] then
        if not post.exists(cgi.pathinfo[1], cgi.pathinfo[2]) then
            html.container.begin();
            io.write("Thread not found.");
            html.container.finish();
            html.finish();
            os.exit();
        end

        html.post.postform(cgi.pathinfo[1], cgi.pathinfo[2]);
        io.write("<hr />");
        html.post.renderthread(cgi.pathinfo[1], cgi.pathinfo[2]);
        io.write("<hr />");

        io.write("<div id='bottom-links' />");
        io.write("<a href='/nano/", cgi.pathinfo[1], "'>[Catalog]</a>");
        io.write("<a href='/nano/", cgi.pathinfo[1], "/", cgi.pathinfo[2], "'>[Update]</a>");
        io.write("<div id='thread-reply'>");
        io.write("<a href='#postform'>[Reply]</a>");
        io.write(#post.threadreplies(cgi.pathinfo[1], cgi.pathinfo[2]), " replies");
        io.write("</div>");
        io.write("</div>");
    else
        html.post.postform(cgi.pathinfo[1], 0);
        html.post.catalog(cgi.pathinfo[1]);
    end

    html.finish();
end
